/*
  Printing Software for CASIO DP-300 (Cygwin)

  2014-02-08 T. Nakagawa

  usage: dp300.exe file.bmp
    file.bmp is a Windows 1-bit/pixel bitmap file.
    The image height must be 120, and the width must be 32~65535.
*/

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define ENVIRON "DP300PORT"
#define COMPORT 2	/* n of COMn port */

#pragma pack(push, 1)

typedef struct BITMAP_FILE_HEADER {
  unsigned short bfType;
  unsigned int bfSize;
  unsigned short bfReserved1;
  unsigned short bfReserved2;
  unsigned int bfOffBits;
} BITMAP_FILE_HEADER;

typedef struct BITMAP_INFO_HEADER {
  unsigned int biSize;
  int biWidth;
  int biHeight;
  unsigned short biPlanes;
  unsigned short biBitCount;
  unsigned int biCompression;
  unsigned int biSizeImage;
  int biXPixPerMeter;
  int biYPixPerMeter;
  unsigned int biClrUsed;
  unsigned int biClrImporant;
} BITMAP_INFO_HEADER;

typedef struct RGB_QUAD {
  unsigned char rgbBlue;
  unsigned char rgbGreen;
  unsigned char rgbRed;
  unsigned char rgbReserved;
} RGB_QUAD;

#pragma pack(pop)

void dmode(int port, int speed) {
  char buf[256];

  sprintf(buf, "mode.com COM%d: BAUD=%d PARITY=N DATA=8 STOP=1 to=on xon=off octs=off odsr=off idsr=off dtr=on rts=off > NUL", port, speed);
  system(buf);
}

int dopen(int port) {
  char device[256];
  int fd;

  sprintf(device, "/dev/ttyS%d", port - 1);
  fd = open(device, O_RDWR);
  if (fd < 0) {
    fprintf(stderr, "ERROR: cannot open the device '%s'\n", device);
    getchar();
    exit(1);
  }
  return fd;
}

void dwrite(int fd, void *data, int size) {
  int n;
  n = write(fd, data, size);
  if (n != size) {
    fprintf(stderr, "ERROR(dwrite): %d != %d\n", n, size);
    getchar();
    exit(1);
  }
}

void expect(int fd, void *data, int size) {
  int i;
  int n;
  unsigned char buf;

  for (i = 0; i < size; i++) {
    n = read(fd, &buf, 1);
    if (n != 1) {
      fprintf(stderr, "ERROR(expect): expected=%02x, actual=\n", ((unsigned char *)data)[i]);
      getchar();
      exit(1);
    }
    if (buf != ((unsigned char *)data)[i]) {
      fprintf(stderr, "ERROR: expected=%02x, actual=%02x\n", ((unsigned char *)data)[i], buf);
      getchar();
      exit(1);
    }
  }
}

int main(int argc, char **argv) {
  int i, j;
  char *bmpfile;
  int port;
  FILE *fp;
  BITMAP_FILE_HEADER file_header;
  BITMAP_INFO_HEADER info_header;
  RGB_QUAD rgb_quad[2];
  int fd;
  unsigned char *image;
  int height;
  int width;
  int block;	/* Image width represented by bytes and aligned to 4 bytes. */
  unsigned char data_size[2];
  unsigned char *data_image;
  unsigned char data_image_chksum;
  unsigned char data_width[2];
  unsigned char data_width_chksum;
  double gray0, gray1;

  if (argc != 2) {
    fprintf(stderr, "usage: %s <*.bmp file>\n", argv[0]);
    return 1;
  }
  bmpfile = argv[1];
  fprintf(stderr, "Image file: %s\n", bmpfile);

  /* Read a bitmap image. */

  fp = fopen(bmpfile, "rb");
  if (fp == NULL) {
    fprintf(stderr, "ERROR: cannot open the image file.\n");
    getchar();
    return 1;
  }
  /* File header. */
  if (fread(&file_header, sizeof(BITMAP_FILE_HEADER), 1, fp) != 1) {
    fprintf(stderr, "ERROR: cannot read the file header.\n");
    getchar();
    return 1;
  }
  if (file_header.bfType != 0x4d42) {
    fprintf(stderr, "ERROR: wrong bfType.\n");
    getchar();
    return 1;
  }
  /* Info header. */
  if (fread(&info_header, sizeof(BITMAP_INFO_HEADER), 1, fp) != 1) {
    fprintf(stderr, "ERROR: cannot read the info header size.\n");
    getchar();
    return 1;
  }
  fprintf(stderr, "Image size: %dx%d\n", info_header.biWidth, info_header.biHeight);
  if (info_header.biSize != 40) {
    fprintf(stderr, "ERROR: wrong info header size (%u).\n", info_header.biSize);
    getchar();
    return 1;
  }
  if (info_header.biBitCount != 1) {
    fprintf(stderr, "ERROR: image bit count is not equal to 1 (%d).\n", info_header.biBitCount);
    getchar();
    return 1;
  }
  if (info_header.biCompression != 0) {
    fprintf(stderr, "ERROR: wrong image compression (%u).\n", info_header.biCompression);
    getchar();
    return 1;
  }
  if (info_header.biHeight != 120) {
    fprintf(stderr, "ERROR: image height is not equal to 120 (%d).\n", info_header.biHeight);
    getchar();
    return 1;
  }
  if (info_header.biWidth < 32) {
    fprintf(stderr, "ERROR: image width is less than 32 (%d).\n", info_header.biWidth);
    getchar();
    return 1;
  }
  if (info_header.biWidth > 65535) {
    fprintf(stderr, "ERROR: image width is larger than 65535 (%d).\n", info_header.biWidth);
    getchar();
    return 1;
  }
  /* Color palette. */
  if (fread(rgb_quad, sizeof(RGB_QUAD), 2, fp) != 2) {
    fprintf(stderr, "ERROR: cannot read the info header.\n");
    getchar();
    return 1;
  }
  /* Bitmap image. */
  height = info_header.biHeight;
  width = info_header.biWidth + ((info_header.biWidth % 8 != 0) ? (8 - info_header.biWidth % 8) : 0);
  block = (info_header.biWidth + ((info_header.biWidth % 32 != 0) ? (32 - info_header.biWidth % 32) : 0)) / 8;
  image = malloc(sizeof(unsigned char) * block * height);
  if (fread(image, sizeof(unsigned char), block * height, fp) != block * height) {
    fprintf(stderr, "ERROR: cannot read image data.\n");
    getchar();
    return 1;
  }
  fclose(fp);

  /* Preapare data. */

  data_width[0] = width / 256;
  data_width[1] = width % 256;
  data_width_chksum = 0;
  data_width_chksum += 0x47;
  data_width_chksum += 0x42;
  data_width_chksum += data_width[0];
  data_width_chksum += data_width[1];
  data_width_chksum += 0x00;
  data_width_chksum += 0x78;
  data_width_chksum = 255 - data_width_chksum;
  data_image = malloc(sizeof(unsigned char) * width / 8 * height);
  for (i = 0; i < height; i++) {
    for (j = 0; j < width / 8; j++) {
      data_image[i * width / 8 + j] = image[(height - 1 - i) * block + j];
    }
  }
  /* Reverse the bitmap if necessary. */
  gray0 = 0.298912 * rgb_quad[0].rgbRed + 0.586611 * rgb_quad[0].rgbGreen + 0.114478 * rgb_quad[0].rgbBlue;
  gray1 = 0.298912 * rgb_quad[1].rgbRed + 0.586611 * rgb_quad[1].rgbGreen + 0.114478 * rgb_quad[1].rgbBlue;
  if (gray0 < gray1) {
    for (i = 0; i < width / 8 * height; i++) data_image[i] ^= 0xff;
  }
  /* Clear the padding. */
  if (info_header.biWidth % 8 != 0) {
    unsigned char mask = 0;
    for (i = 0; i < info_header.biWidth % 8; i++) {
      mask >>= 1;
      mask |= 0x80;
    }
    for (i = 0; i < height; i++) {
      data_image[i * width / 8 + width / 8 - 1] &= mask;
    }
  }

  /* Output to the printer */

  port = (getenv(ENVIRON) == NULL) ? COMPORT : atoi(getenv(ENVIRON));
  fprintf(stderr, "COM port: COM%d\n", port);
  fprintf(stderr, "Printing");

  fprintf(stderr, ".");
  dmode(port, 9600);
  fd = dopen(port);
  dwrite(fd, "\x04", 1);
  dwrite(fd, "\x05", 1);
  expect(fd, "\x06", 1);
  dwrite(fd, "\x43\x42\x31\x31\x35", 5);	/* 115200 bps */
  expect(fd, "\xe3", 1);
  dwrite(fd, "\x06", 1);
  close(fd);

  fprintf(stderr, ".");
  dmode(port, 115200);
  fd = dopen(port);
  dwrite(fd, "\x05", 1);
  expect(fd, "\x06", 1);
  dwrite(fd, "\x05", 1);
  expect(fd, "\x06", 1);
  dwrite(fd, "\x50\x57", 2);
  expect(fd, "\x58", 1);
  dwrite(fd, "\x06", 1);
  expect(fd, "\x3f", 1);
  dwrite(fd, "\x50\x53", 2);
  expect(fd, "\x5c", 1);
  dwrite(fd, "\x06", 1);
  expect(fd, "\x50\x45", 2);
  dwrite(fd, "\x51\x45", 2);
  expect(fd, "\x69", 1);
  dwrite(fd, "\x06", 1);
  close(fd);

  fprintf(stderr, ".");
  dmode(port, 9600);
  fd = dopen(port);
  dwrite(fd, "\x04", 1);
  dwrite(fd, "\x05", 1);
  expect(fd, "\x06", 1);
  dwrite(fd, "\x43\x42\x31\x31\x35", 5);	/* 115200 bps */
  expect(fd, "\xe3", 1);
  dwrite(fd, "\x06", 1);
  close(fd);

  fprintf(stderr, ".");
  dmode(port, 115200);
  fd = dopen(port);
  dwrite(fd, "\x05", 1);
  expect(fd, "\x06", 1);
  dwrite(fd, "\x05", 1);
  expect(fd, "\x06", 1);
  dwrite(fd, "\x47\x42", 2);
  dwrite(fd, data_width, 2);
  dwrite(fd, "\x00\x78", 2);	/* Image height=0x0078=120 */
  expect(fd, (char *)&data_width_chksum, 1);
  dwrite(fd, "\x06", 1);
  dwrite(fd, "\x47\x43\x43\x00\x00\x00\x10\x00\x10", 9);
  expect(fd, "\x12", 1);
  dwrite(fd, "\x06", 1);
  dwrite(fd, "\x47\x51\x00\x00\x00", 5);
  expect(fd, "\x67", 1);
  dwrite(fd, "\x06", 1);
  dwrite(fd, "\x47\x49", 2);
  expect(fd, "\x6f", 1);
  dwrite(fd, "\x06", 1);
  expect(fd, "\x12", 1);
  /* Send the image data by splitting to 0x0600 byte chunks */
  for (i = 0; i * 0x0600 < width / 8 * height; i++) {
    int len;
    len = width / 8 * height - i * 0x0600;
    if (len > 0x0600) len = 0x0600;

    /* Prepare the data. */
    data_size[0] = len / 256;
    data_size[1] = len % 256;
    data_image_chksum = 0;
    data_image_chksum += data_size[0];
    data_image_chksum += data_size[1];
    for (j = i * 0x0600; j < i * 0x0600 + len; j++) data_image_chksum += data_image[j];
    data_image_chksum += 0x17;
    data_image_chksum = 255 - data_image_chksum;
    /* Output to the printer. */
    dwrite(fd, "\x02", 1);
    dwrite(fd, data_size, 2);
    dwrite(fd, data_image + i * 0x0600, len);
    dwrite(fd, "\x17", 1);
    dwrite(fd, &data_image_chksum, 1);
    expect(fd, "\x06", 1);
  }
  dwrite(fd, "\x02\x00\x00\x03\xfc", 5);
  expect(fd, "\x06", 1);
  expect(fd, "\x06", 1);
  dwrite(fd, "\x50\x57", 2);
  expect(fd, "\x58", 1);
  dwrite(fd, "\x06", 1);
  expect(fd, "\x3f", 1);
  dwrite(fd, "\x50\x53", 2);
  expect(fd, "\x5c", 1);
  dwrite(fd, "\x06", 1);
  expect(fd, "\x50\x45", 2);
  dwrite(fd, "\x50\x57", 2);
  expect(fd, "\x58", 1);
  dwrite(fd, "\x06", 1);
  expect(fd, "\x3f", 1);
  dwrite(fd, "\x50\x53", 2);
  expect(fd, "\x5c", 1);
  dwrite(fd, "\x06", 1);
  expect(fd, "\x50\x45", 2);
  dwrite(fd, "\x50\x54", 2);
  expect(fd, "\x5b", 1);
  dwrite(fd, "\x06", 1);
  expect(fd, "\x50\x45", 2);
  dwrite(fd, "\x43\x52", 2);
  expect(fd, "\x6a", 1);
  dwrite(fd, "\x06", 1);
  expect(fd, "\x50\x46", 2);
  dwrite(fd, "\x51\x45", 2);
  expect(fd, "\x69", 1);
  dwrite(fd, "\x06", 1);
  close(fd);

  fprintf(stderr, "Done.\n");

  return 0;
}
